@@ -4,17 +4,19 @@ use 5.006;
 use strict;
 use warnings;
 
-our $VERSION = '1.02';
+our $VERSION = '1.03';
 
 use Chorus::Frame;
 
+use constant DEBUG => 0;
+
 =head1 NAME
 
 Chorus::Engine - A very light inference engine combined with the frame model for knowledge representation.
 
 =head1 VERSION
 
-Version 1.02
+Version 1.04
 
 =cut
 
@@ -119,8 +121,8 @@ Version 1.02
 
 =head2 cut()
 
-       Go directly to the next rule (same loop). This will break all nested instanciation loops
-      on _SCOPE of the current rule.
+       Go directly to the next rule (same loop, same agent). This will break all nested instanciation loops
+       on _SCOPE of the current rule. -> GO DIRECTLY TO NEXT RULE (SAME AGENT)
        
            Ex. $agent->addrule(
              _SCOPE => { .. },
@@ -133,7 +135,8 @@ Version 1.02
 
 =head2 last()
 
-       Terminates the current loop on rules. This will force a cut() too.
+       Breaks the current loop (on rules) for the current agent -> GO DIRECTLY TO NEXT AGENT
+       This will force a cut() too.
        
            Ex. $agent->addrule(
              _SCOPE => { .. },
@@ -144,6 +147,32 @@ Version 1.02
            );
 =cut
 
+=head2 replay()
+
+       Restart FROM THE BEGINNING (1st rule) for the CURRENT AGENT. This will force a cut() too.
+       
+           Ex. $agent->addrule(
+             _SCOPE => { .. },
+             _APPLY => sub {
+              if ( .. ) {
+                 $agent->replay();
+              }
+           );
+=cut
+
+=head2 replay_all()
+
+       Restart FROM THE BEGINNING for the FIRST AGENT. This will force a cut() too.
+       
+           Ex. $agent->addrule(
+             _SCOPE => { .. },
+             _APPLY => sub {
+              if ( .. ) {
+                 $agent->replay_all();
+              }
+           );
+=cut
+
 =head2 solved()
 
        Tells the Chorus::Engine to terminate immediately. This will force a last() too
@@ -194,7 +223,7 @@ Version 1.02
            );
 =cut
            
-=head2 sleep()
+=head2 pause()
 
        Disable a Chorus::Engine object until call to wakeup(). In this mode, the method loop() has no effect.
        This method can optimise the application by de-activating a Chorus::Engine object until it has 
@@ -215,7 +244,7 @@ sub reorderRules {
   my ($funcall) = shift;
   return unless $funcall;
   $SELF->{_RULES} = [ sort { &{$funcall}($a,$b) } @{$SELF->{_RULES}} ];
-  $SELF->{_QUEUE} = [];
+  $SELF->replay;
 }
 
 =head2 applyrules()
@@ -223,6 +252,8 @@ sub reorderRules {
   main engine loop (iterates on $SELF->_RULES)
 =cut
 
+use Data::Dumper;
+
 sub applyrules {
 
   sub apply_rec {
@@ -235,37 +266,63 @@ sub applyrules {
        } grep { $_ ne '_KEY'} keys(%{$rule->{_SCOPE}});
     
     my $i = 0;
-    
+        
     my $head = 'JUMP: {' . join("\n", map { $i++; 'foreach my $k' . $i . ' (@{$scope{' . $_ . '}})' . " {\n\t" . '$opt{' . $_ . '}=$k' . $i . ";" 
                }  keys(%scope)) . "\n";
-    my $body = '$res = $rule->get(\'_APPLY\', %opt); last JUMP if ($SELF->{_LAST} or $SELF->{_CUT} or $SELF->BOARD->SOLVED)';
+               
+    # TODO - SET variables from  %opt HERE !!
+    # ..
+     
+    my $body = '$res = $rule->get(\'_APPLY\', %opt); last JUMP if $SELF->{_LAST} or $SELF->{_CUT} or $SELF->{_REPLAY} or $SELF->{_REPLAY_ALL} or $SELF->BOARD->SOLVED or $SELF->BOARD->FAILED';
     my $tail = "\n}" x scalar(keys(%scope)) . '}';
-    eval $head . $body . $tail; warn $@ if $@;
+
+    eval $head . $body . $tail; 
+    if ($@) {
+       warn $@;
+       warn "DEBUG 1 - Rule '$rule->{_ID}' _SCOPE was : " . join(', ', keys(%scope)) . "\n";
+    }
+    
+    warn "DEBUG - Rule '$rule->{_ID}' returned TRUE. : " if $res and DEBUG;
+    
     $stillworking ||= $res;
 
-    delete $SELF->{_CUT} if $SELF->{_CUT};
+    delete $SELF->{_CUT} if $SELF->{_CUT}; # see eval (already processed) on prev line !!
     
-    $SELF->{_QUEUE} = [] if ($SELF->{_LAST} or $SELF->BOARD->SOLVED);
+    $SELF->{_QUEUE} = [] if $SELF->{_LAST} or $SELF->{_REPLAY} or $SELF->{_REPLAY_ALL} or $SELF->BOARD->SOLVED or $SELF->BOARD->FAILED;
     delete $SELF->{_LAST} if $SELF->{_LAST};
     
+    return undef if $SELF->{_REPLAY} or $SELF->{_REPLAY_ALL};
+
+    $SELF->{_SUCCES} ||= $stillworking;
+    
     return $stillworking unless $SELF->{_QUEUE}->[0];
     return apply_rec (shift @{$SELF->{_QUEUE}}, $stillworking);
   }
 
   return undef if $SELF->{_SLEEPING};
-  $SELF->{_QUEUE} = [ @{$SELF->{_RULES}} ];
+  $SELF->{_QUEUE} = [ @{$SELF->{_RULES} || [] } ];
   return apply_rec(shift @{$SELF->{_QUEUE}});  
 }
 
 my $AGENT = Chorus::Frame->new(
-          cut     => sub { $SELF->{_CUT}  = 'Y' },
-          last    => sub { $SELF->{_LAST} = 'Y' },
-          loop    => sub { do {} while(applyrules() and ! $SELF->BOARD->SOLVED) },
-          sleep   => sub { $SELF->{_SLEEPING} = 'Y' },
+
+          cut         => sub { $SELF->{_CUT}        = 'Y' }, # returns true
+          last        => sub { $SELF->{_LAST}       = 'Y' }, # returns true
+          replay      => sub { $SELF->{_REPLAY}     = 'Y' }, # (returned value ignored)
+          replay_all  => sub { $SELF->{_REPLAY_ALL} = 'Y' }, # (returned value ignored)
+          
+          loop    => sub { $SELF->{_SUCCES} = 0; do {} while(applyrules() and (! $SELF->BOARD or ! $SELF->BOARD->SOLVED or ! $SELF->BOARD->FAILED)) },
+
+          solved  => sub { $SELF->BOARD->{SOLVED} = 'Y'; return undef },
+          failed  => sub { $SELF->BOARD->{FAILED} = 'Y'; return undef },
+          
+          pause   => sub { $SELF->{_SLEEPING} = 'Y' },
           wakeup  => sub { $SELF->delete('_SLEEPING')},
+          
           addrule => sub { push @{$SELF->{_RULES}}, Chorus::Frame->new(@_) },
-          reorder => sub { reorderRules(@_); },
-          solved  => sub { $SELF->BOARD->{SOLVED} = 'Y' },
+          reorder => sub { reorderRules(@_) },
+          
+          debug   => sub { $SELF->{_DEBUG} = shift }
 );
 
 =head2 new
@@ -273,7 +330,9 @@ my $AGENT = Chorus::Frame->new(
 =cut
 
 sub new {
+	my $class = shift;
 	return Chorus::Frame->new(
+	  @_,
 	  _ISA    => $AGENT,
 	  _RULES  => [],
 	)
