@@ -4,7 +4,7 @@ use 5.006;
 use strict;
 use warnings;
 
-our $VERSION = '1.02';
+our $VERSION = '1.03';
 
 =head1 NAME
 
@@ -13,7 +13,7 @@ Chorus::Expert - A simple skeleton of application using one or more Chorus::Engi
 
 =head1 VERSION
 
-Version 1.02
+Version 1.04
 
 =cut
 
@@ -55,7 +55,7 @@ Chorus::Expert does 3 simple things :
 use Chorus::Frame; 
 
 my @agents = ();
-my $board  = Chorus::Frame->new();
+my $board  = Chorus::Frame->new(); # shared with $self->BOARS between agents
 
 sub new {
   my $class = shift;
@@ -71,21 +71,24 @@ sub new {
    
    my $xprt = Chorus::Expert->new();
 
-   my $e1 = Chorus::Engine->new();  # inference engine 1
-   my $e2 = Chorus::Engine->new();  # inference engine 2
+   my $e1 = Chorus::Engine->new();    # inference engine 1
+   my $e2 = Chorus::Engine->new();    # inference engine 2
 
-   $xprt->register($e1,$e2);        # $e1 and $2 added to the list of agents
-                                    # providing to all of them a shared attribute named BOARD
+   $xprt->register($e1,$e2);          # $e1 and $2 added to the list of agents
+                                      # providing to all of them a shared attribute named BOARD
    
 =cut
 
 sub register {
-	my $this = shift;
-	$_->set('BOARD', $board) for @_; 
+	my $this = shift;                 # -> @_ ~equiv. @agents
+	$_->set('BOARD', $board) for @_;
+	$_->set('EXPERT', $this) for @_;
 	push @agents, @_;
 	return $this;
 }
 
+# --
+
 =head2 process
 
    Tells the Chorus::Expert object to enter in an infinite loop until one of the engines 
@@ -99,15 +102,51 @@ sub register {
    
 =cut
 
+my $DEBUG = 0;
+
+sub debug { 
+	my ($this, $level) = @_;
+	$DEBUG = $level; 
+}
+
 sub process {
   my ($this, $input) = @_;
-  $board->set('INPUT', $input);
+  $board->set('INPUT', $input);  # $self->BOARD->INPUT is the default INPUT shared betwwen agents
   do { 
+  	   my @processed = ();
   	   for (@agents) {
-  	   $_->loop() unless $board->{SOLVED}; 
+  	   	
+  	   	 if ($_->_LOCK_UNTIL_STABLE ) {
+  	   	    print STDERR "Chorus::Expert - Agent $_->{_IDENT} is tagged with LOCK_UNTIL_STABLE\n" if $DEBUG;
+  	   	    last if grep { $_->_SUCCES } @processed;
+  	   	    print STDERR "Chorus::Expert - None of agents [" . join (',', map { $_->{_IDENT} || 'NO_NAME' } @processed) . "] have succeeded\n" if $DEBUG;
+  	   	 }
+  	   	
+  	   	 do {
+  	   	 	
+  	   	   if ($_->_REPLAY) {
+  	       	 print STDERR "Chorus::Expert - REPLAYING AGENT $_->{_IDENT} NOW.\n" if $DEBUG;
+  	   	 	 $_->delete('_REPLAY');
+  	   	   }
+  	   	   
+  	   	   print STDERR "Chorus::Expert - LOOPING ON AGENT $_->{_IDENT} NOW.\n" if $DEBUG;
+  	       $_->loop() unless $board->SOLVED or $board->FAILED;
+  	       
+  	     } while($_->_REPLAY);
+
+         push @processed, $_;
+         
+  	   	 if ($_->_REPLAY_ALL) {
+  	   	 	print STDERR "Chorus::Expert - WILL REPLAY ALL AGENTS NOW.\n" if $DEBUG;
+  	   	 	$_->delete('_REPLAY_ALL');
+  	   	 	last;
+  	   	 }
        } 
-  } until ($board->{SOLVED});
-  $board->delete('SOLVED');
+  } until ($board->{SOLVED} or $board->{FAILED});
+  
+  ($board->delete('SOLVED'), return 1)     if $board->{SOLVED};
+  ($board->delete('FAILED'), return undef) if $board->{FAILED};
+  
 }
 
 =head1 AUTHOR
