@@ -3,7 +3,7 @@ package Chorus::Frame;
 use 5.006;
 use strict;
 
-our $VERSION = '1.02';
+our $VERSION = '1.03';
 
 =head1 NAME
 
@@ -11,7 +11,7 @@ Chorus::Frame - A short implementation of frames from knowledge representation.
 
 =head1 VERSION
 
-Version 1.02
+Version 1.04
 
 =cut
 
@@ -74,8 +74,8 @@ BEGIN {
   use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
 
   @ISA         = qw(Exporter);
-  @EXPORT      = qw($SELF &fmatch);
-  @EXPORT_OK   = qw();	
+  @EXPORT      = qw($SELF &fmatch REQUIRE_FAILED);
+  @EXPORT_OK   = qw(%FMAP);	
 
   # %EXPORT_TAGS = ( );		# eg: TAG => [ qw!name1 name2! ];
 }
@@ -91,15 +91,16 @@ use constant DEBUG_MEMORY => 0;
 
 use vars qw($AUTOLOAD);
 
-use constant SUCCESS => 1;
-use constant FAILED  => 0;
+use constant SUCCESS        =>  1;
+use constant FAILED         =>  0;
+use constant REQUIRE_FAILED => -1;
 
 use constant VALUATION_ORDER => ('_VALUE', '_DEFAULT', '_NEEDED');
 
 use constant MODE_N => 1;
 use constant MODE_Z => 2;
 
-my $getMode = MODE_N;
+my $getMode = MODE_N;       # DEFAULT IS N !!!
 
 my %REPOSITORY;
 my %FMAP;
@@ -168,7 +169,6 @@ sub popself {
 }
 
 sub expand {
-  
     my ($info, @args) = @_;
     return expand(&$info(@args)) if _isa($info, 'CODE');
     return $info;       
@@ -181,15 +181,17 @@ sub expand {
 
 sub _push {
   my ($this, $slot, @elems) = @_;
-  return $this->{$slot} = [ @elems ] unless exists $this->{$slot};
-  $this->{$slot} = ref($this->{$slot}) eq 'ARRAY' ? [ @{$this->{$slot}}, @elems ] : [ $this->{$slot}, @elems ];  
+  return unless scalar(@elems);
+  $this->{$slot} = [ $this->{$slot} || () ] unless ref($this->{$slot}) eq 'ARRAY'; 
+  unshift @{$this->{$slot}}, @elems;
 }
 
 sub _addInstance {
   my ($this, $instance) = @_;
   my $k = $instance->{_KEY};
+  
   $INSTANCES{$this->{_KEY}}->{$k} = $instance;
-  weaken($INSTANCES{$this->{_KEY}}->{$k}) ; # not counted in garbage collector !
+  weaken($INSTANCES{$this->{_KEY}}->{$k});      # will not increase garbage ref counter to $instance !!
 }
 
 =head2 _inherits
@@ -200,8 +202,12 @@ sub _addInstance {
   
 sub _inherits {
   my ($this, @inherited) = @_;
-  $_->_addInstance($this) for @inherited;
-  $this->_push('_ISA', @inherited); # shoult test if already inherited !
+  my $k = $this->{_KEY};
+  for (grep { ! $INSTANCES{$_->{_KEY}}->{$k} } @inherited) { # clean list
+  	$_->_addInstance($this);
+  	$this->_push('_ISA', $_);
+  }
+  return $this;
 }
 
 sub _removeInstance {
@@ -229,7 +235,7 @@ sub blessToFrame {
     
     $this->{_KEY} = $k;
     $FMAP{$k} = $this;
-    weaken($FMAP{$k}) ; # not counted in garbage collector !
+    weaken($FMAP{$k}); # will not increase garbage ref counter to $this !!
     return $this; 
   }
 
@@ -327,6 +333,8 @@ sub new {
   return blessToFrame({@desc});
 }
 
+# WARN - Should be automatically called by carbage collector EVEN with those 2 remaining references : $INSTANCES{$k} AND $FMAP{$k} !!!
+#
 sub DESTROY { 
   my ($this) = @_;
 
@@ -383,9 +391,9 @@ sub get {
 
     my $res = $this->first($tryValuations,@args);
     return $res if defined($res) and $res->{ret};
-  
+
     if (exists($this->{_ISA})) {
-      my @h = _isa($this->{_ISA}, 'ARRAY') ? map \&expand, @{$this->{_ISA}} : (expand($this->{_ISA}));
+      my @h = _isa($this->{_ISA}, 'ARRAY') ? map { expand($_) || () } @{$this->{_ISA}} : (expand($this->{_ISA}) || ());
       for (@h) { # upper level
         $res = $_->expandInherits($tryValuations,@args);
         return $res if defined($res) and $res->{ret};
@@ -393,7 +401,7 @@ sub get {
     }
     return { ret => FAILED };
   } # expandInherits
-  
+
   sub inherited {
 	my ($this,$slot,@rest) = @_;
 
@@ -404,7 +412,7 @@ sub get {
 	return undef unless $next;
 	return $next->inherited($slot,@rest);
   } 
-	
+
   sub getZ {
   	
   	sub value_Z {
@@ -548,6 +556,8 @@ sub set {
   sub setValue {
     my ($this, $val) = @_;
 
+    return undef if $this->getN('_REQUIRE', $val) == REQUIRE_FAILED;
+    
     $this->getN('_BEFORE', $val); # or return undef;
 
     blessToFrame($val);
@@ -595,7 +605,7 @@ sub set {
     
     return $this->{$nextStep}->setN($followWay, $info); # (keep current context)
     
-  } # setN
+  } # set is setN !!!!
 	
   pushself(shift);
   my %desc = @_;
@@ -603,7 +613,7 @@ sub set {
   my $res;
  
   while(my($k,$val) = each %desc) {
-    $res = $SELF->setN($k, $val);
+    $res = $SELF->setN($k, $val);    # set is setN !!
   }
   
   popself();
@@ -649,7 +659,7 @@ sub fmatch {
         
        my ($res, @dig) = @_;
          return $res unless $dig[0];
-         my @inheriting = map { $_->firstInheriting } @dig;
+         my @inheriting = map { $_->firstInheriting || () } @dig;
          push(@$res, @inheriting);
          return wholeTree($res,@inheriting);
          
@@ -658,7 +668,7 @@ sub fmatch {
     my ($slot) = @_;
     
     my @res = hasSlot($slot);
-    my @inheriting = map { $_->firstInheriting } @res;
+    my @inheriting = map { $_->firstInheriting || () } @res;
   
     push @res, @inheriting;
     return wholeTree(\@res, @inheriting);
@@ -671,7 +681,7 @@ sub fmatch {
 
   return () unless $firstslot;
   
-  my %filter = map { $_->{_KEY} => 'Y' } @{framesProvidingSlot($firstslot)};
+  my %filter = map { $_->{_KEY} => 'Y' || () } @{framesProvidingSlot($firstslot)};
   
   for(@otherslots) {
     %filter = map { $filter{$_->{_KEY}} ? ($_->{_KEY} => 'Y') : () } @{framesProvidingSlot($_)};
@@ -681,7 +691,7 @@ sub fmatch {
     return grep { $filter{$_->{_KEY}} } @{$opts{from}};
   }
   
-  return map { $FMAP{$_} } keys(%filter);
+  return map { $FMAP{$_} || ()} keys(%filter);
   
 } # fmatch
 
